Declare DriverEntry(*DriverObject, *RegistryPath)

; Import kernel function of "ntoskrnl.exe".
Import "ntoskrnl.lib"
  IoCompleteRequest(*IRP, PriorityBoost)
  RtlInitUnicodeString(*UString, *String)
  IoCreateDevice(*DriverObject, DeviceExtensionSize, 
                 *UDeviceName, DeviceType,
                 DeviceCharacteristics, Exclusive, *DeviceObject)
  IoDeleteDevice(*DeviceObject)
  IoCreateSymbolicLink(*SymbolicLinkName, *DeviceName)
  IoDeleteSymbolicLink(*SymbolicLinkName)
EndImport


#IOCTL_READ_PORT_UCHAR	 = $200 
#IOCTL_WRITE_PORT_UCHAR	 = $400

; Entry point in main function the driver.
*EntryPoint = @DriverEntry()
!jmp [p_EntryPoint]

EnableExplicit

; Direct write in port.
Procedure.a InPort(Port.u)
  Protected Result.a=0
  EnableASM
  MOV DX, Port
  IN  al,DX
  MOV Result, al
  DisableASM
  ProcedureReturn Result
EndProcedure

; Direct reading from port.
Procedure OutPort(Port.u, Byte.a)
  EnableASM
  MOV al, Byte
  MOV DX, Port
  OUT DX, al
  DisableASM
EndProcedure

; This procedure, called kernel manager,
; calling WinAPI functions DeviceIoControl_().
Procedure DeviceIoControl(*DeviceObject.DEVICE_OBJECT, *pIrp.IRP)
  Protected ntStatus, *Point, *Stack.IO_STACK_LOCATION
  Protected InBuffersize, OutBuffersize, CtrlBuff
  Protected Port.u, Byte.a, InByte.a, Code
  
  *Stack = *pIrp\Tail\Overlay\CurrentStackLocation
  
  ; This is the "dwIoControlCode", sent messages of WinAPI functions DeviceIoControl_().
  Code = *Stack\Parameters\DeviceIoControl\IoControlCode
  
  ; This is the "nInBufferSize", sent messages of WinAPI functions DeviceIoControl_().
  InBuffersize = *Stack\Parameters\DeviceIoControl\InputBufferLength
  
  ; This is the "nOutBufferSize", sent messages of WinAPI functions DeviceIoControl_().
  OutBuffersize = *Stack\Parameters\DeviceIoControl\OutputBufferLength
  
  ; This is the "lpInBuffer", sent messages of WinAPI functions DeviceIoControl_().
  CtrlBuff   = PeekI(*pIrp\SystemBuffer)
  Port = CtrlBuff & $FFFF
  
  ntStatus = #STATUS_SUCCESS ; This value returns WinAPI functions DeviceIoControl_().
  
  Select Code
    Case #IOCTL_READ_PORT_UCHAR ; Reading from a port.
      InByte=InPort(Port)  ; 
      PokeA(*pIrp\SystemBuffer, InByte) ; Read a byte from the port is placed in a buffer.
      *pIrp\IoStatus\Information = 1    ; In buffer, write 1 byte.
      
    Case #IOCTL_WRITE_PORT_UCHAR ; Writing in Port.
      Byte = (CtrlBuff>>16)&255
      OutPort(Port, Byte)
      *pIrp\IoStatus\Information = 0  ; In buffer, write 0 byte.
      
    Default                          ; The command is not supported.
      ntStatus = #STATUS_UNSUCCESSFUL
      *pIrp\IoStatus\Information = 0  ; In buffer, write 0 byte.
  EndSelect
  
  ; Message processed.
  ; More information. http://msdn.microsoft.com/en-us/library/windows/hardware/ff548343(v=vs.85).aspx
  *pIrp\IoStatus\Status = ntStatus
  IoCompleteRequest(*pIrp, #IO_NO_INCREMENT)
  
  ProcedureReturn ntStatus
EndProcedure


Procedure CreateDispatch(*DeviceObject.DEVICE_OBJECT, *pIrp.IRP)
  
  *pIrp\IoStatus\Information = 0
  *pIrp\IoStatus\Status = #STATUS_SUCCESS
  IoCompleteRequest(*pIrp, #IO_NO_INCREMENT)
  
  ProcedureReturn #STATUS_SUCCESS
EndProcedure

; Procedure for unloading driver.
Procedure UnloadDriver(*DriverObject.DRIVER_OBJECT)
  Protected uniDOSString.UNICODE_STRING
  
  RtlInitUnicodeString(@uniDOSString, ?DosDevices)
  IoDeleteSymbolicLink (@uniDOSString)
  IoDeleteDevice(*DriverObject\DeviceObject)
  
EndProcedure

; Procedure initialize driver.
Procedure DriverEntry(*DriverObject.DRIVER_OBJECT, *RegistryPath.UNICODE_STRING)
  Protected deviceObject.DEVICE_OBJECT
  Protected uniNameString.UNICODE_STRING
  Protected uniDOSString.UNICODE_STRING
  Protected status
  
  RtlInitUnicodeString(@uniNameString, ?Device)
  RtlInitUnicodeString(@uniDOSString, ?DosDevices)
  
  status = IoCreateDevice(*DriverObject, 0, @uniNameString, #FILE_DEVICE_UNKNOWN, 0, #False, @deviceObject)
  If status <> #STATUS_SUCCESS
    ProcedureReturn status
  EndIf
  
  status = IoCreateSymbolicLink(@uniDOSString, @uniNameString)
  If status <> #STATUS_SUCCESS
    IoDeleteDevice(@deviceObject) ; We need to remove an object!
                                  ; In mode kernel, of system will
                                  ; not do it even after unloading the driver!
    ProcedureReturn status
  EndIf
  
  
  ; Registration procedures.
  *DriverObject\DriverUnload = @UnloadDriver() ; Procedure for unloading drivers
  *DriverObject\MajorFunction[#IRP_MJ_CREATE] = @CreateDispatch() ; Called when WinAPI, CreateFile_().
  *DriverObject\MajorFunction[#IRP_MJ_CLOSE] = @CreateDispatch() ; Called when WinAPI, CloseHandle_().
  *DriverObject\MajorFunction[#IRP_MJ_DEVICE_CONTROL] = @DeviceIoControl() ; ; Called when WinAPI, DeviceIoControl_().
  
  ProcedureReturn #STATUS_SUCCESS
EndProcedure



; Name of the driver in system.
DataSection
  Device:
  !du '\Device\pbdriverio', 0, 0
  
  DosDevices:
  !du '\DosDevices\pbdriverio', 0, 0
EndDataSection



; IDE Options = PureBasic 5.10 Beta 8 (Windows - x86)
; CursorPosition = 157
; FirstLine = 117
; Folding = --
; Executable = pbdriverio.sys
; DisableDebugger
; CompileSourceDirectory
; IncludeVersionInfo
; VersionField0 = 1,2,0,0
; VersionField1 = 1,2,0,0
; VersionField2 = PB_Dr
; VersionField3 = pbdriverio
; VersionField4 = 1.2
; VersionField5 = 1.2
; VersionField6 =      .
; VersionField7 = pbdriverio
; VersionField8 = pbdriverio.sys
; VersionField13 = pure-basic@yandex.ru
; VersionField14 = http://www.pure-basic.narod.ru/
; VersionField15 = VOS_NT
; VersionField16 = VFT_DRV
; VersionField17 = 0419 Russian